#include <stdlib.h>
#include "hashtbl.h"
#include "hash.h"

/* Q9: from TP9 */
typedef struct cell *list;

struct cell {
  key k;
  value v;
  list next;
};


/* Q10: a hash table is an array of lists of size N 
   we use calloc so that it is initialized with the empty list
   (== NULL [== 0])
*/
void hashtbl_init(hashtbl* h, int N) {
  *h = calloc(N, sizeof (list));
  set_cardinal(N);
}

/* Q11: list_add from TP9 */
void list_add(key k, value v, list *l) {
  list r = malloc(sizeof (struct cell));
  r->k = k;
  r->v = v;
  r->next = *l;
  *l = r;
}

void hashtbl_add(key k, value v, hashtbl* h) {
  list *r = *h;  // type cast to have an array of list and not of void
  int hk = hash(k);
  list_add(k, v, &r[hk]);
}

/* Q12: list_find from TP9 */
value list_find(key k, list l) {
  while (l != NULL) {
    if (l->k == k)
      return l->v;
    l = l->next;
  }
  return -1;
}

value hashtbl_find(key k, hashtbl h) {
  list *r = h;
  int hk = hash(k);
  return list_find(k, r[hk]);
}

/* Q13: list_remove from TP9 */
void list_remove(key k, list *l) {
  if (*l == NULL) return;
  list curr = *l;
  if (curr->k == k) {
    *l = curr->next;
    free(curr);
    return;
  }
  list prec = curr;
  curr = curr->next;
  while (curr != NULL) {
    if (curr->k == k) {
      prec->next = curr->next;
      free(curr);
      return;
    }
    prec = curr;
    curr = curr->next;
  }
}

void hashtbl_remove(key k, hashtbl *h) {
  list *r = *h;
  int hk = hash(k);
  list_remove(k, &r[hk]);
}

/* Q14: first version commented */
/*
void hashtbl_replace(key k, value v, hashtbl *h) {
  hashtbl_remove(k, h);
  hashtbl_add(k, v, h);
}
*/

void hashtbl_replace(key k, value v, hashtbl *h) {
  list *r = *h;
  int hk = hash(k);
  list l = r[hk];
  while (l != NULL && l->k != k)
    l = l->next;
  if (l == NULL)
    list_add(k, v, &r[hk]);
  else
    l->v = v;
}
